DESTINATION_HEADER

uniform sampler2D s_curve_0;
uniform float u_opacity;

uniform float u_hue_shift_radians;
uniform float u_sat_plus_one;
uniform float u_lightness;

uniform sampler2D s_mask;
varying vec2 v_mask_uv;


vec3 HSLfromRGB(vec3 color);
vec3 RGBfromHSL(vec3 hsl);

void main()
{
	vec4 dst = DESTINATION_FETCH;

#ifdef OLDWAY
	mat3 m = mat3										// sRGB to Ycc
	(												    // following CR's derivation from RGB -> Y
		vec3(0.2126729 ,  0.5       , -0.05730415),
		vec3(0.7151522 , -0.4541646 , -0.19269587),
		vec3(0.072175  , -0.01944736,  0.25)
	);
	mat3 m1 = mat3										// Ycc to sRGB
	(												    // following CR's derivation from RGB -> Y
		 vec3(0.95844787,  1.01235669,  1.),
		 vec3(1.5746542 , -0.46827273,  0.),
		 vec3(-0.15421241, -0.32869406,  3.7113)
	);

	vec3 ycc = m * dst.rgb;
	ycc.r = texture2D(s_curve_0, vec2(ycc.r, 0.5)).a; 	// lookup Y

	float theta = atan(ycc.b, ycc.g);
	float radius = length(ycc.gb);

	theta += u_hue_shift_radians;
	radius *= u_sat_plus_one;

	ycc.g = radius * cos(theta);
	ycc.b = radius * sin(theta);

	vec4 value;
	value.rgb = m1 * ycc;
    value.a = dst.a;
#else
	// New compatible math that uses RGB -> HSV -> RGB
	// like Photoshop.
	//
	// Important: the old technique also supported an arbitrary
	// luminosity curve -- the "Plus" in "hueSatPlus". The new
	// math does not support the luma curve since we are only trying
	// match the Photoshop "Hue/Saturation" adjustment layer.
	//

	vec3 hsl = HSLfromRGB(dst.rgb);

	float shift = u_hue_shift_radians / (3.14159 * 2.0);
	hsl.x -= shift;
	if (hsl.x < 0.0)
		hsl.x += 1.0;
	if (hsl.x > 1.0)
		hsl.x -= 1.0;

	hsl.y *= u_sat_plus_one;
	if (hsl.y > 1.0)
		hsl.y = 1.0;

	vec4 value;
	value.rgb = RGBfromHSL(hsl);
	value.a = dst.a;

#endif

    if (u_lightness < 0.0)
    {
        vec3 black = vec3(0.0, 0.0, 0.0);
        value.rgb = mix(black, value.rgb, u_lightness + 1.0);
    }
    else if (u_lightness > 0.0)
    {
        vec3 white = vec3(1.0, 1.0, 1.0);
        value.rgb = mix(value.rgb, white, u_lightness);
    }

	float mask = texture2D(s_mask, v_mask_uv).a;
	float factor = u_opacity * mask;

	gl_FragColor = mix(dst, value, factor);
}


// Search "HSL Image Adjustements on GPU" for information regarding
// optimal HSL math on GPUs.
///

vec3 HSLfromRGB(vec3 color)
{
	float minc = min(color.r, min(color.g, color.b));
	float maxc = max(color.r, max(color.g, color.b));
	float delta = maxc - minc;

	if (delta == 0.0)
		return vec3(0.0, 0.0, minc);

	float lum = (minc + maxc) * 0.5;
	float hue = 0.0;
	float sat = 0.0;

	if (lum > 0.0 && lum < 1.0) {
		float mul = (lum < 0.5) ? (lum) : (1.0-lum);
		sat = delta / (mul * 2.0);
	}

	vec3 masks = vec3(
		(maxc == color.r && maxc != color.g) ? 1.0 : 0.0,
		(maxc == color.g && maxc != color.b) ? 1.0 : 0.0,
		(maxc == color.b && maxc != color.r) ? 1.0 : 0.0
	);

	vec3 adds = vec3(
					  ((color.g - color.b) / delta),
				2.0 + ((color.b - color.r) / delta),
				4.0 + ((color.r - color.g) / delta)
	);

	hue += dot(adds, masks);
	hue *= (1.0/6.0);			// hue [0, 6.0] -> [0.0, 1.0]
	if (hue < 0.0)
		hue += 1.0;

	return vec3(hue, sat, lum);
}

vec3 RGBfromHSL(vec3 hsl)
{
	float one_third = 1.0 / 3.0;
	float two_third = 2.0 / 3.0;
	float hue_sixth = 6.0;

	float hue = hsl.x;
	float sat = hsl.y;
	float lum = hsl.z;

	vec3 xt;

	if (hue < one_third)
	{
		xt.r = hue_sixth * (one_third - hue);
		xt.g = hue_sixth * hue;
		xt.b = 0.0;
	}
	else if (hue < two_third)
	{
		xt.r = 0.0;
		xt.g = hue_sixth * (two_third - hue);
		xt.b = hue_sixth * (hue - one_third);
	}
	else
	{
		xt.r = hue_sixth * (hue - two_third);
		xt.g = 0.0;
		xt.b = hue_sixth * (1.0 - hue);
	}

	xt.r = min(xt.r, 1.0);
	xt.g = min(xt.g, 1.0);
	xt.b = min(xt.b, 1.0);

	float sat2   =  2.0 * sat;
	float satinv =  1.0 - sat;
	float luminv =  1.0 - lum;
	float lum2m1 = (2.0 * lum) - 1.0;
	vec3  ct     = (sat2 * xt) + satinv;

	vec3 rgb;
	if (lum >= 0.5)
		rgb = (luminv * ct) + lum2m1;
	else
		rgb = lum * ct;

	return rgb;
}